import datetime
import decimal
import io
import os
import pathlib
import posixpath
import warnings
from decimal import Decimal
from functools import partial
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Tuple, Union
from xml.dom.minidom import parse

from django.core.exceptions import SuspiciousFileOperation
from django.core.files import File
from django.core.files.storage import FileSystemStorage, Storage
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.fields.files import FieldFile
from django.db.models.utils import make_model_tuple
from django.utils.functional import SimpleLazyObject
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from PIL import Image
from pyexpat import ExpatError
from variations.typing import Size
from variations.utils import prepare_image, replace_extension

from .. import exceptions, helpers, signals
from ..conf import settings
from ..files import VariationFile
from ..typing import FileLike
from ..utils import cached_method, checksum
from ..variations import PaperVariation
from .mixins import FileFieldProxyMixin, FileProxyMixin
from .query import ResourceQuerySet

try:
    from django.core.files.utils import validate_file_name
except ImportError:
    # new in Django 3.1.10
    def validate_file_name(name, allow_relative_path=False):
        if os.path.basename(name) in {'', '.', '..'}:
            raise SuspiciousFileOperation("Could not derive file name from '%s'" % name)

        if allow_relative_path:
            # Use PurePosixPath() because this branch is checked only in
            # FileField.generate_filename() where all file paths are expected to be
            # Unix style (with forward slashes).
            path = pathlib.PurePosixPath(name)
            if path.is_absolute() or '..' in path.parts:
                raise SuspiciousFileOperation(
                    "Detected path traversal attempt in '%s'" % name
                )
        elif name != os.path.basename(name):
            raise SuspiciousFileOperation("File name '%s' includes path elements" % name)
        return name

__all__ = [
    "NoPermissionsMetaBase",
    "ResourceBaseMeta",
    "Resource",
    "FileResource",
    "FileFieldResource",
    "SVGFileResourceMixin",
    "ImageFileResourceMixin",
    "VariationFile",
    "VersatileImageResourceMixin",
]


class Permissions(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("upload", "Can upload files"),
            ("change", "Can change files"),
            ("delete", "Can delete files"),
        )


class NoPermissionsMetaBase:
    """
    Отменяет создание автоматических объектов Permission у наследников класса.
    """
    def __new__(cls, name, bases, attrs, **kwargs):
        meta = attrs.pop("Meta", None)
        if meta is None:
            meta = type("Meta", (), {"default_permissions": ()})
        else:
            meta_attrs = meta.__dict__.copy()
            meta_attrs.setdefault("default_permissions", ())
            meta = type("Meta", meta.__bases__, meta_attrs)
        attrs["Meta"] = meta

        return super().__new__(cls, name, bases, attrs, **kwargs)


class OverridableParentLink:
    """
    Приём, позволяющий переопределить OneToOne-связь между моделями при наследовании
    от промежуточной абстрактной модели.

    По умолчанию, при наследовании от абстрактной модели, которая в свою очередь
    унаследована от конкретной (concrete), попытка переопределния OneToOne-связи
    не замещает поле по умолчанию, а добавляет второе.

    Типичный пример использования - удаление обратной связи (related_name)
    в пользовательским классе, унаследованным от `CollectionItemBase` с целью
    избежать ошибки 'reverse accessor clashes ...' в тех случаях, когда имя
    пользовательской модели не уникально.

    См. пример `ImageItem` в папке `tests/examples/collections/custom_models/models.py`.

    Ссылка:
    https://docs.djangoproject.com/en/3.2/topics/db/models/#specifying-the-parent-link-field
    """
    def __new__(cls, name, bases, attrs, **kwargs):
        parents = [b for b in bases if isinstance(b, ModelBase)]

        parent_links = {}
        for base in reversed(parents):
            # Conceptually equivalent to `if base is Model`.
            if not hasattr(base, "_meta"):
                continue

            # Locate OneToOneField instances.
            for field in base._meta.local_fields:
                if isinstance(field, models.OneToOneField) and field.remote_field.parent_link:
                    key = make_model_tuple(field.remote_field.model)
                    parent_links.setdefault(key, []).append(field)

        for fieldname, field in list(attrs.items()):
            if isinstance(field, models.OneToOneField) and field.remote_field.parent_link:
                key = make_model_tuple(field.remote_field.model)
                if key in parent_links:
                    inherited_field = parent_links[key].pop()
                    if fieldname != inherited_field.name:
                        # force delete inherited field
                        attrs[inherited_field.name] = None

        return super().__new__(cls, name, bases, attrs, **kwargs)


class ResourceOptions:
    """
    Опции для моделей ресурса.
    Добавлить опции можно через вложенный класс `ResourceMeta`:

        class MyResource(Resource):
            class ResourceMeta:
                required_fields = []
    """
    # Поля, необходимые для инициализации модели. В случаях, когда
    # набор запрашиваемых из БД полей ограничивается (например, при использовании
    # методов `Queryset.only()` и `Queryset.defer()`), поля, перечисленные
    # в этом параметре, автоматически добавляются к списку извлекаемых из БД полей.
    required_fields = []

    def __init__(self, opts=None, **kwargs):
        # Override defaults with options provided
        if opts:
            opts = list(opts.__dict__.items())
        else:
            opts = []
        opts.extend(list(kwargs.items()))

        for key, value in opts:
            if not key.startswith("__"):
                setattr(self, key, value)

    def __iter__(self):
        return ((k, v) for k, v in self.__dict__.items() if k[0] != "_")


class ResourceBaseMeta(NoPermissionsMetaBase, OverridableParentLink, models.base.ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        attr_meta = attrs.pop("ResourceMeta", None)
        if attr_meta is None:
            attr_meta = type("ResourceMeta", (), {})

        local_options = frozenset(dir(attr_meta))
        parents = [b for b in bases if isinstance(b, ModelBase)]

        # extend ResourceMeta from base classes
        for base in parents:
            if hasattr(base, "_resource_meta"):
                for key, value in base._resource_meta:
                    if key not in local_options:
                        setattr(attr_meta, key, value)

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        new_class.add_to_class("_resource_meta", ResourceOptions(attr_meta))
        return new_class


class ResourceBase(models.Model, metaclass=ResourceBaseMeta):
    """
    Базовый класс ресурса.
    """
    created_at = models.DateTimeField(
        _("created at"),
        default=now,
        editable=False
    )
    modified_at = models.DateTimeField(
        _("changed at"),
        auto_now=True,
        editable=False
    )

    objects = ResourceQuerySet.as_manager()

    class Meta:
        abstract = True

    def __repr__(self):
        return "{} #{}".format(type(self).__name__, self.pk)

    def as_dict(self) -> Dict[str, Any]:
        """
        Представление объекта в виде словаря, конвертируемого в JSON.
        Служит для формирования виджета файла без перезагрузки страницы.
        """
        return {
            "id": self.pk,
            "created": self.created_at.isoformat() if self.created_at else None,
            "modified": self.modified_at.isoformat() if self.modified_at else None,
        }


class Resource(ResourceBase):
    """
    Ресурс.

    TODO: после переноса миксины BacklinkModelMixin класс остался пустым.
    """

    class Meta(ResourceBase.Meta):
        abstract = True


class FileResource(FileProxyMixin, Resource):
    """
    Подкласс ресурса, представляющего файл.
    """

    resource_name = models.CharField(
        _("resource name"),
        max_length=255,
        editable=False,
        help_text=_("Human-readable resource name"),
    )
    extension = models.CharField(
        _("extension"),
        max_length=32,
        editable=False,
        help_text=_("Lowercase, without leading dot"),
    )
    size = models.PositiveIntegerField(
        _("size"),
        default=0,
        editable=False
    )
    checksum = models.CharField(
        _("checksum"),
        max_length=64,
        editable=False,
    )
    uploaded_at = models.DateTimeField(
        _("uploaded at"),
        default=now,
        editable=False
    )

    class Meta(Resource.Meta):
        abstract = True

    def __str__(self):
        return self.get_caption()

    def __repr__(self):
        return "{}('{}')".format(type(self).__name__, self.name)

    @property
    def basename(self):
        warnings.warn(
            "'basename' is deprecated in favor of 'resource_name'",
            DeprecationWarning,
            stacklevel=2
        )
        return self.resource_name

    @basename.setter
    def basename(self, value):
        warnings.warn(
            "'basename' is deprecated in favor of 'resource_name'",
            DeprecationWarning,
            stacklevel=2
        )
        self.resource_name = value

    @property
    def name(self) -> str:
        """
        Полное имя загруженного файла.
        Это имя включает относительный путь, суффикс и расширение.
        """
        raise NotImplementedError

    def get_caption(self) -> str:
        """
        Человекопонятное имя файла.
        Не содержит суффикса, которое может быть добавлено файловым хранилищем.
        """
        if self.extension:
            return "{}.{}".format(self.resource_name, self.extension)
        return self.resource_name

    def get_file_size(self) -> int:
        """
        Получение размера загруженного файла.
        """
        raise NotImplementedError

    def file_exists(self) -> bool:
        """
        Проверка существования файла.
        """
        raise NotImplementedError

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "name": self.resource_name,
            "extension": self.extension,
            "caption": self.get_caption(),
            "size": self.size,
            "uploaded": self.uploaded_at.isoformat() if self.uploaded_at else None
        }

    def update_checksum(self, file: FileLike = None) -> bool:
        file = file or self.get_file()
        old_checksum = self.checksum
        new_checksum = checksum(file)
        if new_checksum and new_checksum != old_checksum:
            signals.checksum_update.send(
                sender=type(self),
                instance=self,
                checksum=new_checksum
            )
            self.checksum = new_checksum
        return old_checksum != new_checksum

    def attach(self, file: Union[str, Path, FileLike], name: str = None, **options):
        """
        Присоединение файла к экземпляру ресурса.
        Этот метод является обёрткой. Фактическое сохранение файла
        происходит в методе `_attach()`.

        Если на данном этапе обнаруживается, что переданный файл не может
        быть представлен этой моделью, необходимо вызвать исключение
        UnsupportedResource.
        """
        if not name:
            if isinstance(file, str):
                name = file
            elif isinstance(file, Path):
                name = str(file)
            else:
                name = getattr(file, "name", None)

        # prevent path traversal
        if name:
            name = os.path.basename(name)

        # convert to `django.core.files.File` instance
        need_to_close = False
        if isinstance(file, (str, Path)):
            need_to_close = True
            file = File(open(file, "rb"), name=name)
        elif isinstance(file, File):
            file.name = name
        else:
            file = File(file, name=name)

        # reset file position before prepare
        if file.seekable():
            file.seek(0)

        prepared_file = self._prepare_file(file, **options)

        signals.pre_attach_file.send(
            sender=type(self), instance=self, file=prepared_file, options=options
        )

        # reset file position before upload
        if prepared_file.seekable():
            prepared_file.seek(0)

        attach_result = self._attach(prepared_file, **options)

        self.size = self.get_file_size()
        self.uploaded_at = now()
        self.modified_at = now()

        # Рассчет хэша от входного файла, а не от загруженного,
        # т.к. в случае Cloudinary, это приведет к избыточному
        # скачиванию файла из облачного хранилища.
        if prepared_file.seekable():
            prepared_file.seek(0)
            self.update_checksum(prepared_file)
        else:
            self.update_checksum()

        signals.post_attach_file.send(
            sender=type(self),
            instance=self,
            file=self.get_file(),
            options=options,
            response=attach_result,
        )

        if need_to_close:
            if not file.closed:
                file.close()

    def attach_file(self, *args, **kwargs):
        warnings.warn(
            "attach_file() is deprecated in favor of attach()",
            DeprecationWarning,
            stacklevel=2
        )
        return self.attach(*args, **kwargs)

    def _prepare_file(self, file: File, **options) -> File:
        """
        Подготовка файла к присоединению к модели.
        """
        return file

    def _attach(self, file: File, **options):
        raise NotImplementedError

    def _attach_file(self, *args, **kwargs):
        warnings.warn(
            "_attach_file() is deprecated in favor of _attach()",
            DeprecationWarning,
            stacklevel=2
        )
        return self._attach(*args, **kwargs)

    def rename(self, new_name: str, **options):
        """
        Переименование файла.
        Этот метод является обёрткой. Фактическое переименование файла
        происходит в методе `_rename()`.
        """
        self._require_file()

        if not self.file_exists():
            raise FileNotFoundError(self.name)

        old_name = self.name

        signals.pre_rename_file.send(
            sender=type(self),
            instance=self,
            old_name=old_name,
            new_name=new_name,
            options=options,
        )

        response = self._rename(new_name, **options)

        self.modified_at = now()

        signals.post_rename_file.send(
            sender=type(self),
            instance=self,
            old_name=old_name,
            new_name=new_name,
            options=options,
            response=response,
        )

    def rename_file(self, *args, **kwargs):
        warnings.warn(
            "rename_file() is deprecated in favor of rename()",
            DeprecationWarning,
            stacklevel=2
        )
        return self.rename(*args, **kwargs)

    def _rename(self, new_name: str, **options):
        raise NotImplementedError

    def _rename_file(self, *args, **kwargs):
        warnings.warn(
            "_rename_file() is deprecated in favor of _rename()",
            DeprecationWarning,
            stacklevel=2
        )
        return self._rename(*args, **kwargs)

    def delete_file(self, **options):
        """
        Удаление файла.
        Этот метод является обёрткой. Фактическое удаление файла
        происходит в методе `_delete_file()`.
        """
        signals.pre_delete_file.send(sender=type(self), instance=self, options=options)
        response = self._delete_file(**options)
        signals.post_delete_file.send(sender=type(self), instance=self, options=options, response=response)

    def _delete_file(self, **options):
        raise NotImplementedError


class FileFieldResource(FileFieldProxyMixin, FileResource):
    """
    Подкласс файлового ресурса, доступ к которому осуществляется через Django Storage.
    """

    class Meta(FileResource.Meta):
        abstract = True

    def _require_file(self):
        file = self.get_file()
        if not file:
            file_field = self.get_file_field()
            raise ValueError(_("The '%s' attribute has no file associated with it.") % file_field.name)

    def _attach(self, file: File, **options):
        self.get_file().save(file.name, file, save=False)
        self.resource_name = helpers.get_filename(file.name)
        self.extension = helpers.get_extension(self.name)

    def _rename(self, new_name: str, **options):
        # копирование файла, чтобы не вредить кэшированию
        file = self.get_file()
        with file.open() as fp:
            file.save(new_name, fp, save=False)

        self.resource_name = helpers.get_filename(new_name)
        self.extension = helpers.get_extension(self.name)

    def _delete_file(self, **options):
        self.get_file().delete(save=False)

    @property
    def name(self) -> str:
        self._require_file()
        return self.get_file().name

    @classmethod
    def get_file_field(cls):
        """
        Получение файлового поля модели.
        """
        raise NotImplementedError

    def get_file_folder(self) -> str:
        """
        Возвращает путь к папке, в которую будет сохранен файл.
        """
        raise NotImplementedError

    def get_file_storage(self) -> Storage:
        raise NotImplementedError

    def generate_filename(self, filename: str) -> str:
        """
        Формирование каталога и имени файла при сохранении.
        """
        dirname = self.get_file_folder()
        dirname = datetime.datetime.now().strftime(dirname)
        filename = posixpath.join(dirname, filename)
        filename = validate_file_name(filename, allow_relative_path=True)

        storage = self.get_file_storage()
        return storage.generate_filename(filename)

    def get_file(self) -> FieldFile:
        raise NotImplementedError

    def set_file(self, value):
        field = self.get_file_field()
        setattr(self, field.attname, value)

    def get_file_size(self) -> int:
        self._require_file()
        return self.get_file().size

    def get_file_url(self) -> str:
        warnings.warn(
            "get_file_url() is deprecated in favor of 'url' property",
            DeprecationWarning,
            stacklevel=2
        )
        return self.url

    def file_exists(self) -> bool:
        file = self.get_file()
        if not file:
            return False
        return file.storage.exists(file.name)

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "url": self.url,
        }


class SVGFileResourceMixin(models.Model):
    """
    Подкласс файлового ресурса для SVG-изображений.
    """

    title = models.CharField(
        _("title"),
        max_length=255,
        blank=True,
        help_text=_(
            "The title is being used as a tooltip when the user hovers the mouse over the image"
        ),
    )
    description = models.TextField(
        _("description"),
        blank=True,
        help_text=_(
            "This text will be used by screen readers, search engines, or when the image cannot be loaded"
        ),
    )
    width = models.DecimalField(
        _("width"),
        max_digits=10,
        decimal_places=4,
        default=0,
        editable=False
    )
    height = models.DecimalField(
        _("height"),
        max_digits=10,
        decimal_places=4,
        default=0,
        editable=False
    )

    class Meta:
        abstract = True

    @property
    def width_display(self):
        width = str(self.width)
        return width.rstrip("0").rstrip(".") if "." in width else width

    @property
    def height_display(self):
        height = str(self.height)
        return height.rstrip("0").rstrip(".") if "." in height else height

    @property
    def ratio(self) -> Decimal:
        """
        Width-to-height (W/H) ratio.
        """
        return Decimal(str(round(self.width / self.height, 8)))

    @property
    def hw_ratio(self) -> Decimal:
        """
        Height-to-width (H/W) ratio.
        """
        return Decimal(str(round(self.height / self.width, 8)))

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),  # noqa
            "width": self.width_display,
            "height": self.height_display,
            "title": self.title,
            "description": self.description,
        }

    def _prepare_file(self, file: File, **options) -> File:
        try:
            dom = parse(file)
        except ExpatError:
            raise exceptions.UnsupportedResource(
                _("File `%s` is not an svg image") % file.name
            )

        root = dom.documentElement
        if root.tagName.lower() != "svg":
            raise exceptions.UnsupportedResource(
                _("File `%s` is not an svg image") % file.name
            )

        width = root.getAttribute("width")
        height = root.getAttribute("height")
        view_box = root.getAttribute("viewBox")
        view_box = view_box.split(" ") if view_box else []

        if not width and len(view_box) == 4:
            width = view_box[2]

        if not height and len(view_box) == 4:
            height = view_box[3]

        try:
            self.width = round(decimal.Decimal(width), 4)
        except decimal.InvalidOperation:
            self.width = 0

        try:
            self.height = round(decimal.Decimal(height), 4)
        except decimal.InvalidOperation:
            self.height = 0

        return super()._prepare_file(file, **options)  # noqa: F821


class ImageFileResourceMixin(models.Model):
    """
    Подкласс файлового ресурса для изображений
    """

    title = models.CharField(
        _("title"),
        max_length=255,
        blank=True,
        help_text=_(
            "The title is being used as a tooltip when the user hovers the mouse over the image"
        ),
    )
    description = models.TextField(
        _("description"),
        blank=True,
        help_text=_(
            "This text will be used by screen readers, search engines, or when the image cannot be loaded"
        ),
    )
    width = models.PositiveSmallIntegerField(
        _("width"),
        default=0,
        editable=False
    )
    height = models.PositiveSmallIntegerField(
        _("height"),
        default=0,
        editable=False
    )
    cropregion = models.CharField(
        _("crop region"),
        max_length=24,
        blank=True,
        editable=False
    )

    class Meta:
        abstract = True

    @property
    def ratio(self) -> Decimal:
        """
        Width-to-height (W/H) ratio.
        """
        return Decimal(str(round(self.width / self.height, 8)))

    @property
    def hw_ratio(self) -> Decimal:
        """
        Height-to-width (H/W) ratio.
        """
        return Decimal(str(round(self.height / self.width, 8)))

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),  # noqa
            "width": self.width,
            "height": self.height,
            "cropregion": self.cropregion,
            "title": self.title,
            "description": self.description,
        }

    def _prepare_file(self, file: File, **options) -> File:
        try:
            image = Image.open(file)
        except OSError:
            raise exceptions.UnsupportedResource(
                _("File `%s` is not an image") % file.name
            )
        else:
            self.width, self.height = image.size

            # format extension
            file.name = replace_extension(file.name, format=image.format)
            root, ext = os.path.splitext(file.name)
            ext = ext.lstrip(".").lower()
            file.name = ".".join([root, ext])

        return super()._prepare_file(file, **options)  # noqa: F821


class VersatileImageResourceMixin(ImageFileResourceMixin):
    """
    Подкласс файлового ресурса для вариативного изображения.
    """

    # класс файла вариации
    variation_class = VariationFile

    # флаг, запускающий нарезку вариаций после сохранения экземпляра модели в БД
    need_recut = False

    class Meta(ImageFileResourceMixin.Meta):
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reset_variation_files()

    def __getattr__(self, item):
        # реализация-заглушка, чтобы PyCharm не ругался на атрибуты-вариации
        raise AttributeError(
            "{!r} object has no attribute {!r}".format(self.__class__.__name__, item)
        )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.need_recut:
            self.need_recut = False
            if settings.RQ_ENABLED:
                self.recut_async()
            else:
                self.recut()

    def _reset_variation_files(self):
        """
        Обновление файлов вариаций в текущем экземпляре.
        """

        # Очистка кэша метода variation_files()
        if hasattr(self, self.variation_files.cache_key):
            delattr(self, self.variation_files.cache_key)

        try:
            for vname in self.get_variations():
                self.__dict__[vname] = SimpleLazyObject(partial(self.get_variation_file, vname))
        except exceptions.CollectionModelNotFoundError:
            pass

    @cached_method("_variation_files_cache")
    def variation_files(self) -> Tuple[Tuple[str, VariationFile]]:
        if not self.get_file():
            raise cached_method.Bypass(tuple())

        return tuple(
            (vname, self.get_variation_file(vname))
            for vname in self.get_variations()
        )

    def get_variation_file(self, variation_name: str) -> VariationFile:
        return self.variation_class(instance=self, variation_name=variation_name)

    def get_variations(self) -> Dict[str, PaperVariation]:
        raise NotImplementedError

    def attach(self, file: FileLike, name: str = None, **options):
        super().attach(file, name=name, **options)  # noqa: F821
        self.need_recut = True
        self._reset_variation_files()

    def rename(self, new_name: str, **options):
        super().rename(new_name, **options)  # noqa: F821
        self.need_recut = True
        self._reset_variation_files()

    def delete_file(self, **options):
        super().delete_file(**options)  # noqa: F821
        self.delete_variations()

    def delete_variations(self):
        for vname, vfile in self.variation_files():
            vfile.delete()

        self._reset_variation_files()

    def recut(self, names: Iterable[str] = ()):
        """
        Нарезка вариаций.
        Можно указать имена конкретных вариаций в параметре `names`.
        """
        if not self.file_exists():
            raise FileNotFoundError(self.name)

        file = self.get_file()
        with file.open() as source:
            img = Image.open(source)
            draft_size = self.calculate_max_size(img.size)
            img = prepare_image(img, draft_size=draft_size)

            for vname, variation in self.get_variations().items():
                if names and vname not in names:
                    continue

                image = variation.process(img)
                self._save_variation(vname, variation, image)

                signals.variation_created.send(
                    sender=type(self),
                    instance=self,
                    name=vname
                )

    def calculate_max_size(self, source_size: Size) -> Optional[Tuple[int, int]]:
        """
        Вычисление максимально возможных значений ширины и высоты изображения
        среди всех вариаций, чтобы передать их в Image.draft().
        """
        max_width = 0
        max_height = 0
        for name, variation in self.get_variations().items():
            size = variation.get_output_size(source_size)
            max_width = max(max_width, size[0])
            max_height = max(max_height, size[1])
        if max_width and max_height:
            return max_width, max_height

    def _save_variation(self, name: str, variation: PaperVariation, image: Image):
        """
        Запись изображения в файловое хранилище
        """
        variation_file = self.get_variation_file(name)

        if isinstance(variation_file.storage, FileSystemStorage):
            with variation_file.open("wb") as fp:
                variation.save(image, fp)
        else:
            # Не все Storage-классы позволяют записывать контент с помощью вызовов
            # `open()` и `write()`. Для них нужно использовать метод `_save()`.
            with io.BytesIO() as buffer:
                variation.save(image, buffer)
                content = File(buffer, name=variation_file.name)
                variation_file.storage._save(variation_file.name, content)

    def recut_async(self, names: Iterable[str] = ()):
        """
        Добавление задачи нарезки вариаций в django-rq.
        """
        from django_rq.queues import get_queue

        queue = get_queue(settings.RQ_QUEUE_NAME)
        queue.enqueue_call(
            self._recut_task,
            kwargs={
                "app_label": self._meta.app_label,
                "model_name": self._meta.model_name,
                "object_id": self.pk,
                "using": self._state.db,
                "names": names,
            },
        )

    @classmethod
    def _recut_task(
        cls,
        app_label: str,
        model_name: str,
        object_id: int,
        using: str,
        names: Iterable[str],
    ):
        """
        Задача для django-rq.
        Вызывает `recut()` экземпляра в отдельном процессе.
        """
        instance = helpers.get_instance(app_label, model_name, object_id, using=using)
        instance.recut(names)



from django.contrib import admin
from django.urls import path

from .. import views
from ..models.collection import Collection
from .base import ResourceAdminBase


class CollectionAdminBase(ResourceAdminBase):
    create_collection_view_class = views.collection.CreateCollectionView
    delete_collection_view_class = views.collection.DeleteCollectionView
    upload_item_view_class = views.collection.UploadFileView
    delete_item_view_class = views.collection.DeleteFileView
    change_item_view_class = views.collection.ChangeFileView
    sort_items_view_class = views.collection.SortItemsView

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        urlpatterns = [
            path(
                "create/",
                self.admin_site.admin_view(self.create_collection_view_class.as_view()),
                name="%s_%s_create" % info,
            ),
            path(
                "delete/",
                self.admin_site.admin_view(self.delete_collection_view_class.as_view()),
                name="%s_%s_delete" % info,
            ),
            path(
                "upload-item/",
                self.admin_site.admin_view(self.upload_item_view_class.as_view()),
                name="%s_%s_upload_item" % info,
            ),
            path(
                "delete-item/",
                self.admin_site.admin_view(self.delete_item_view_class.as_view()),
                name="%s_%s_delete_item" % info,
            ),
            path(
                "change-item/",
                self.admin_site.admin_view(self.change_item_view_class.as_view()),
                name="%s_%s_change_item" % info,
            ),
            path(
                "sort-items/",
                self.admin_site.admin_view(self.sort_items_view_class.as_view()),
                name="%s_%s_sort_items" % info,
            ),
        ]
        return urlpatterns


@admin.register(Collection)
class CollectionAdmin(CollectionAdminBase):
    pass



# Generated by Django 3.2.6 on 2021-09-06 15:05

import django.core.files.storage
from django.db import migrations, models
import paper_uploads.models.fields.image
import paper_uploads.models.utils


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0002_auto_20210407_0627'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fileitem',
            name='file',
            field=models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to=paper_uploads.models.utils.generate_filename, verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='imageitem',
            name='file',
            field=paper_uploads.models.fields.image.VariationalFileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to=paper_uploads.models.utils.generate_filename, verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='svgitem',
            name='file',
            field=models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to=paper_uploads.models.utils.generate_filename, verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='uploadedfile',
            name='file',
            field=models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to=paper_uploads.models.utils.generate_filename, verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='uploadedimage',
            name='file',
            field=paper_uploads.models.fields.image.VariationalFileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to=paper_uploads.models.utils.generate_filename, verbose_name='file'),
        ),
    ]



from copy import deepcopy
from typing import Dict, Iterable

from django import conf
from django.utils.module_loading import import_string

DEFAULTS = {
    "STORAGE": "django.core.files.storage.FileSystemStorage",
    "STORAGE_OPTIONS": {},
    "FILES_UPLOAD_TO": "files/%Y/%m/%d",
    "IMAGES_UPLOAD_TO": "images/%Y/%m/%d",
    "COLLECTION_FILES_UPLOAD_TO": "collections/files/%Y/%m/%d",
    "COLLECTION_IMAGES_UPLOAD_TO": "collections/images/%Y/%m/%d",
    "COLLECTION_ITEM_PREVIEW_WIDTH": 180,
    "COLLECTION_ITEM_PREVIEW_HEIGHT": 135,

    "RQ_ENABLED": False,
    "RQ_QUEUE_NAME": "default",
    "VARIATION_DEFAULTS": None,
}

# Иконки для файлов в галерее
FILE_ICON_DEFAULT = "unknown"
FILE_ICON_OVERRIDES = {
    "3gp": "video",
    "aac": "audio",
    "docx": "doc",
    "flac": "audio",
    "flv": "video",
    "gz": "archive",
    "jpeg": "jpg",
    "m4a": "audio",
    "m4v": "video",
    "mkv": "video",
    "mov": "video",
    "mpeg": "video",
    "ogg": "audio",
    "ogv": "video",
    "xlsx": "xls",
    "xz": "archive",
    "wav": "audio",
    "webm": "video",
    "wma": "audio",
    "wmv": "video",
}


class Settings:
    """
    A settings object, that allows API settings to be accessed as properties.
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings: Dict = None, defaults: Dict = None, import_strings: Iterable = None):
        self.user_settings = user_settings
        self.defaults = defaults
        self.import_strings = import_strings

    def __getstate__(self):
        return {
            "user_settings": self.user_settings,
            "defaults": self.defaults,
            "import_strings": self.import_strings,
        }

    def __getattr__(self, name):
        if name.startswith("prepare_"):
            raise AttributeError(name)

        self.__dict__[name] = value = self._get_value(name)
        return value

    def __copy__(self):
        clone = type(self)()
        clone.__dict__.update(
            user_settings=self.user_settings,
            defaults=self.defaults,
            import_strings=self.import_strings,
        )
        return clone

    def __deepcopy__(self, memo):
        clone = type(self)()
        memo[id(self)] = clone
        clone.user_settings = deepcopy(self.user_settings, memo)
        clone.defaults = deepcopy(self.defaults, memo)
        clone.import_strings = deepcopy(self.import_strings, memo)
        return clone

    def _get_value(self, name):
        value = self._fetch_value(name)

        # Coerce import strings into classes
        if name in self.import_strings:
            return self._perform_import(value, name)

        # Prepare value
        method_name = "prepare_{}".format(name.lower())
        prepare_method = getattr(self, method_name, None)
        if prepare_method is not None:
            value = prepare_method(value)

        return value

    def _fetch_value(self, name):
        try:
            # Check if present in user settings
            return self.user_settings[name]
        except KeyError:
            pass

        try:
            # Fall back to defaults
            return self.defaults[name]
        except KeyError:
            pass

        raise AttributeError("Invalid setting: '%s'" % name)

    @classmethod
    def _perform_import(cls, val, setting_name):
        """
        If the given setting is a string import notation,
        then perform the necessary import or imports.
        """
        if val is None:
            return
        elif isinstance(val, str):
            return cls._import_from_string(val, setting_name)
        elif isinstance(val, (list, tuple)):
            return [cls._import_from_string(item, setting_name) for item in val]
        return val

    @staticmethod
    def _import_from_string(val, setting_name):
        """
        Attempt to import a class from a string representation.
        """
        try:
            return import_string(val)
        except ImportError as e:
            msg = "Could not import '%s' for API setting '%s'. %s: %s." % (
                val,
                setting_name,
                e.__class__.__name__,
                e,
            )
            raise ImportError(msg)


settings = Settings(
    user_settings=getattr(conf.settings, "PAPER_UPLOADS", {}),
    defaults=DEFAULTS,
    import_strings={"STORAGE"}
)


IMAGE_ITEM_VARIATIONS = dict(
    admin_preview=dict(
        size=(
            settings.COLLECTION_ITEM_PREVIEW_WIDTH,
            settings.COLLECTION_ITEM_PREVIEW_HEIGHT
        ),
        format="jpeg",
        versions={"webp", "2x"},
        jpeg=dict(
            quality=75
        ),
        webp=dict(
            quality=65
        ),
    ),
)



from django import forms


class ResourceFieldBase(forms.ModelChoiceField):
    def __init__(
        self,
        *args,
        owner_app_label=None,
        owner_model_name=None,
        owner_fieldname=None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.widget.owner_app_label = owner_app_label
        self.widget.owner_model_name = owner_model_name
        self.widget.owner_fieldname = owner_fieldname


class FileResourceFieldBase(ResourceFieldBase):
    def __init__(
        self,
        *args,
        configuration=None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.widget.configuration = configuration



from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS, migrations, transaction
from django.db.migrations.operations.base import Operation
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.timezone import now

from .. import exceptions
from ..models import CollectionItemBase
from ..models.fields.base import ResourceFieldBase
from .classes import ExtendableMigration


@receiver(post_delete, sender=CollectionItemBase)
def on_delete_collection_item(sender, instance, **kwargs):
    """
    Обновление поля `modified_at` коллекции при удалении элемента,
    чтобы метод `get_last_modified()` возвращал корректные данные.
    """
    if instance.collection_id and instance.collection_content_type_id:
        try:
            collection_cls = instance.get_collection_class()
        except exceptions.CollectionModelNotFoundError:
            return

        collection_cls.objects.filter(
            pk=instance.collection_id
        ).update(
            modified_at=now()
        )


def inject_operations(
    plan=None, apps=global_apps, using=DEFAULT_DB_ALIAS, **kwargs
):
    if plan is None:
        return

    for migration, backward in plan:
        if migration.name == "0001_initial":
            continue

        PaperMigration(
            migration=migration,
            backward=backward,
            apps=apps,
            using=using
        ).iterate()


class PaperMigration(ExtendableMigration):
    def process(self, operation: Operation, **kwargs):
        if isinstance(operation, migrations.RenameField):
            self.insert_after(
                RenameOwnerField(
                    self.migration.app_label,
                    operation.model_name,
                    operation.old_name,
                    operation.new_name,
                )
            )
        elif isinstance(operation, migrations.RenameModel):
            self.insert_after(
                RenameOwnerModel(
                    self.migration.app_label,
                    operation.old_name_lower,
                    operation.new_name_lower,
                )
            )


class RenameOwnerField(migrations.RunPython):
    """
    При переименовании поля, ссылающегося на ресурс, необходимо исправить
    значение owner_fieldname во всех связанных экземплярах.
    """

    def __init__(self, app_label, model_name, old_name, new_name):
        self.app_label = app_label
        self.model_name = model_name
        self.old_name = old_name
        self.new_name = new_name
        super().__init__(self.forward, self.backward)

    def forward(self, apps, schema_editor):
        self._rename(apps, schema_editor, self.old_name, self.new_name)

    def backward(self, apps, schema_editor):
        self._rename(apps, schema_editor, self.new_name, self.old_name)

    def _rename(self, apps, schema_editor, old_name, new_name):
        using = schema_editor.connection.alias
        model = apps.get_model(self.app_label, self.model_name)

        # Текущая миграция выполняется после миграции переименования,
        # поэтому целевое поле уже имеет новое имя.
        field = model._meta.get_field(new_name)

        if isinstance(field, ResourceFieldBase):
            with transaction.atomic(using=using):
                field.related_model._base_manager.db_manager(using).filter(
                    owner_app_label=self.app_label,
                    owner_model_name=self.model_name,
                    owner_fieldname=old_name,
                ).update(
                    owner_fieldname=new_name
                )


class RenameOwnerModel(migrations.RunPython):
    """
    При переименовании модели, в которой есть поля, ссылающиеся на ресурсы,
    необходимо исправить значение owner_model_name во всех связанных экземплярах.
    """

    def __init__(self, app_label, old_name, new_name):
        self.app_label = app_label
        self.old_name = old_name
        self.new_name = new_name
        super().__init__(self.forward, self.backward)

    def forward(self, apps, schema_editor):
        self._rename(apps, schema_editor, self.old_name, self.new_name)

    def backward(self, apps, schema_editor):
        self._rename(apps, schema_editor, self.new_name, self.old_name)

    def _rename(self, apps, schema_editor, old_name, new_name):
        using = schema_editor.connection.alias

        # Текущая миграция выполняется после миграции переименования,
        # поэтому модель уже имеет новое имя.
        model = apps.get_model(self.app_label, new_name)

        with transaction.atomic(using=using):
            for field in model._meta.fields:
                if isinstance(field, ResourceFieldBase):
                    field.related_model._base_manager.db_manager(using).filter(
                        owner_app_label=self.app_label,
                        owner_model_name=old_name,
                        owner_fieldname=field.name,
                    ).update(
                        owner_model_name=new_name
                    )



# Generated by Django 4.0.1 on 2022-02-16 10:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0002_rename_basename_fileitem_resource_name_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fileitem',
            name='resource_name',
            field=models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name'),
        ),
        migrations.AlterField(
            model_name='imageitem',
            name='resource_name',
            field=models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name'),
        ),
        migrations.AlterField(
            model_name='svgitem',
            name='resource_name',
            field=models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name'),
        ),
        migrations.AlterField(
            model_name='uploadedfile',
            name='resource_name',
            field=models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name'),
        ),
        migrations.AlterField(
            model_name='uploadedimage',
            name='resource_name',
            field=models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name'),
        ),
    ]



from typing import Any, Type, TypeVar

from django.contrib.contenttypes.models import ContentType
from django.db import models

from .. import exceptions

T = TypeVar("T")


def get_model_class(content_type_id: int, base_class: Type[T]) -> Type[T]:
    """
    Получение класса модели загружаемого файла по ContentType ID.
    """
    try:
        content_type = ContentType.objects.get_for_id(content_type_id)
    except ContentType.DoesNotExist:
        raise exceptions.InvalidContentType(content_type_id)

    model_class = content_type.model_class()
    if issubclass(model_class, base_class):
        return model_class

    raise exceptions.InvalidContentType(content_type_id)


def get_instance(model_class: Type[models.Model], pk: Any) -> models.Model:
    """
    Получение экземпляра модели загружаемого файла.
    """
    try:
        return model_class._default_manager.get(pk=pk)
    except (ValueError, TypeError):
        raise exceptions.InvalidObjectId(pk)



from typing import Dict, Type

from django.db import models
from django.db.models.fields import Field

from ..helpers import build_variations
from ..models.base import Resource, VersatileImageResourceMixin
from ..models.collection import CollectionBase, CollectionItemBase
from ..models.fields.collection import CollectionItem
from ..variations import PaperVariation


def is_variations_allowed(model: Type[models.Model]) -> bool:
    """
    Возвращает True, если модель представляет из себя класс изображения с вариациями.
    """
    return issubclass(model, VersatileImageResourceMixin)


def is_collection(model: Type[models.Model]) -> bool:
    """
    Возвращает True, если модель - коллекция.
    """
    return issubclass(model, CollectionBase)


def get_field_variations(field: Field) -> Dict[str, PaperVariation]:
    """
    Получение списка вариаций для поля.
    """
    variation_config = getattr(field, "variations", {}).copy()
    return build_variations(variation_config)


def get_collection_variations(
    collection_cls: Type[CollectionBase], item_type_field: CollectionItem
) -> Dict[str, PaperVariation]:
    """
    Получение списка вариаций для коллекции.
    """
    image_model = item_type_field.model
    variation_config = image_model.get_variation_config(collection_cls, item_type_field)
    return build_variations(variation_config)


def is_resource_field(field: Field):
    """
    Возвращает True, если поле ссылается на Resource (но не на элемент коллекции).
    """
    return (
        field.is_relation
        and field.concrete
        and not field.auto_created
        and issubclass(field.related_model, Resource)
        and not issubclass(field.related_model, CollectionItemBase)
    )


def includes_variations(model: Type[models.Model]) -> bool:
    """
    Проверка наличия вариаций в рамках модели.

    Если модель - коллекция, проверяется наличие хотя бы одного класса элементов,
    поддерживающего вариации изображений.
    Если модель - обыкновенная, проверяется наличие поля ImageField (или аналогичного),
    которое ссылается на ресурс, поддерживающий вариативность изображений.
    """
    if is_collection(model):
        return any(
            is_variations_allowed(field.model)
            for field in model.item_types.values()
        )
    else:
        return any(
            is_resource_field(field) and is_variations_allowed(field.related_model)
            for field in model._meta.get_fields(include_hidden=True)
        )



from django.db.models.query import QuerySet
from polymorphic.managers import PolymorphicManager as DefaultPolymorphicManager
from polymorphic.query import PolymorphicQuerySet


class ResourceQuerysetMixin:
    @property
    def required_fields(self):
        return self.model._resource_meta.required_fields

    def only(self, *fields):
        new_fields = tuple(set(fields).union(self.required_fields))
        return super().only(*new_fields)

    def defer(self, *fields):
        new_fields = tuple(set(fields).difference(self.required_fields))
        return super().defer(*new_fields)


class ResourceQuerySet(ResourceQuerysetMixin, QuerySet):
    pass


class PolymorphicResourceQuerySet(ResourceQuerysetMixin, PolymorphicQuerySet):
    pass


class PolymorphicResourceManager(DefaultPolymorphicManager):
    queryset_class = PolymorphicResourceQuerySet



# Generated by Django 3.2.9 on 2021-11-06 10:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0003_auto_20210906_1505'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='fileitem',
            options={'default_permissions': ()},
        ),
        migrations.AlterModelOptions(
            name='imageitem',
            options={'default_permissions': ()},
        ),
        migrations.AlterModelOptions(
            name='svgitem',
            options={'default_permissions': ()},
        ),
    ]



from ... import forms
from .base import FileResourceFieldBase


class SVGFileField(FileResourceFieldBase):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("to", "paper_uploads.UploadedSVGFile")
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        return super().formfield(**{"form_class": forms.FileField, **kwargs})



from typing import Any, List, cast

from django.contrib.contenttypes.fields import GenericRelation
from django.core import checks
from django.db import DEFAULT_DB_ALIAS
from django.db.models import Field
from django.utils.functional import cached_property

from ... import forms
from ...helpers import _get_item_types
from .base import FileResourceFieldBase


class ContentItemRelation(GenericRelation):
    """
    FIX: cascade delete polymorphic
    https://github.com/django-polymorphic/django-polymorphic/issues/34
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def bulk_related_objects(self, objs, using=DEFAULT_DB_ALIAS):
        return super().bulk_related_objects(objs).non_polymorphic()


class CollectionField(FileResourceFieldBase):
    def __init__(self, to, **kwargs):
        kwargs["blank"] = True
        super().__init__(to=to, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "blank" in kwargs:
            del kwargs["blank"]
        return name, path, args, kwargs

    def _check_relation_class(self, base, error_message):
        from ...models.collection import CollectionBase

        rel_is_string = isinstance(self.remote_field.model, str)
        if rel_is_string:
            return []

        model_name = (
            self.remote_field.model
            if rel_is_string
            else self.remote_field.model._meta.object_name
        )
        if not issubclass(self.remote_field.model, CollectionBase):
            return [
                checks.Error(
                    "Field defines a relation with model '%s', which is not "
                    "subclass of CollectionBase" % model_name,
                    obj=self,
                )
            ]
        return []

    def formfield(self, **kwargs):
        return super().formfield(**{"form_class": forms.CollectionField, **kwargs})


class CollectionItem:
    """
    Поле для подключения классов элементов коллекции.
    Может использоваться только в подклассах CollectionBase.
    """

    default_validators = []  # type: List[Any]

    def __init__(self, to, name=None, validators=(), options=None):
        self.name = name

        try:
            to._meta.model_name
        except AttributeError:
            assert isinstance(to, str), (
                "%s(%r) is invalid. First parameter to %s must be a model"
                % (self.__class__.__name__, to, self.__class__.__name__,)
            )

        self.model = to
        self.options = options or {}
        self._validators = list(validators)

    def check(self, **kwargs):
        return [
            *Field._check_field_name(cast(Field, self)),
            *Field._check_validators(cast(Field, self)),
        ]

    @cached_property
    def validators(self):
        return [*self.default_validators, *self._validators]

    def contribute_to_class(self, cls, name, **kwargs):
        self.name = self.name or name
        item_types = _get_item_types(cls)
        if item_types is not None:
            item_types[self.name] = self



# Generated by Django 3.2.9 on 2021-11-16 07:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0005_alter_collectionitembase_index_together'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='fileitem',
            name='owner_app_label',
        ),
        migrations.RemoveField(
            model_name='fileitem',
            name='owner_fieldname',
        ),
        migrations.RemoveField(
            model_name='fileitem',
            name='owner_model_name',
        ),
        migrations.RemoveField(
            model_name='imageitem',
            name='owner_app_label',
        ),
        migrations.RemoveField(
            model_name='imageitem',
            name='owner_fieldname',
        ),
        migrations.RemoveField(
            model_name='imageitem',
            name='owner_model_name',
        ),
        migrations.RemoveField(
            model_name='svgitem',
            name='owner_app_label',
        ),
        migrations.RemoveField(
            model_name='svgitem',
            name='owner_fieldname',
        ),
        migrations.RemoveField(
            model_name='svgitem',
            name='owner_model_name',
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-24 12:26

from django.db import migrations


def forward_func(apps, schema_editor):
    ContentType = apps.get_model("contenttypes", "ContentType")
    Collection = apps.get_model("paper_uploads", "Collection")
    CollectionItemBase = apps.get_model("paper_uploads", "CollectionItemBase")
    db_alias = schema_editor.connection.alias

    collection_updates = []
    for collection in Collection._default_manager.using(db_alias).all().iterator():
        collection_class = apps.get_model(
            collection.collection_content_type.app_label,
            collection.collection_content_type.model
        )
        concrete_collection_class = collection_class._meta.concrete_model
        concrete_content_type = ContentType._default_manager.get_for_model(concrete_collection_class)
        collection.concrete_collection_content_type_id = concrete_content_type.id
        collection_updates.append(collection)

    Collection._default_manager.bulk_update(collection_updates, ["concrete_collection_content_type_id"])

    item_updates = []
    for item in CollectionItemBase._default_manager.using(db_alias).all().iterator():
        collection_class = apps.get_model(
            item.collection_content_type.app_label,
            item.collection_content_type.model
        )
        concrete_collection_class = collection_class._meta.concrete_model
        concrete_content_type = ContentType._default_manager.get_for_model(concrete_collection_class)
        item.concrete_collection_content_type_id = concrete_content_type.id
        item_updates.append(item)

    CollectionItemBase._default_manager.bulk_update(item_updates, ["concrete_collection_content_type_id"])


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ('paper_uploads', '0009_collection_concrete_collection_content_type_and_more'),
    ]

    operations = [
        migrations.RunPython(forward_func, reverse_func),
    ]



# Generated by Django 3.2.9 on 2021-11-16 05:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('paper_uploads', '0004_auto_20211106_1004'),
    ]

    operations = [
        migrations.AlterIndexTogether(
            name='collectionitembase',
            index_together={('collection_id', 'collection_content_type')},
        ),
    ]



from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from .. import helpers


class Command(BaseCommand):
    help = """
    Удаление экземпляров коллекций, в которых нет ни одного элемента.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "--min-age",
            type=int,
            default=30 * 24 * 3600,
            help="Minimum instance age in seconds to look for",
        )

    def handle(self, *args, **options):
        helpers.remove_empty_collections(
            min_age=options["min_age"],
            database=options["database"]
        )



from typing import Any, Dict

from django.core.files.storage import Storage
from django.db.models.fields.files import FieldFile
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..helpers import build_variations
from ..storage import default_storage
from ..utils import cached_method, filesizeformat
from ..variations import PaperVariation
from .base import FileFieldResource, VersatileImageResourceMixin
from .fields import VariationalFileField
from .mixins import BacklinkModelMixin, EditableResourceMixin


class UploadedImageBase(VersatileImageResourceMixin, BacklinkModelMixin, EditableResourceMixin, FileFieldResource):
    change_form_class = "paper_uploads.forms.dialogs.image.ChangeUploadedImageDialog"

    file = VariationalFileField(
        _("file"),
        max_length=255,
    )

    class Meta(FileFieldResource.Meta):
        abstract = True
        verbose_name = _("image")
        verbose_name_plural = _("images")

    class ResourceMeta:
        required_fields = [
            "owner_app_label",
            "owner_model_name",
            "owner_fieldname",
        ]

    def get_file_folder(self) -> str:
        owner_field = self.get_owner_field()
        return getattr(owner_field, "upload_to", "") or settings.IMAGES_UPLOAD_TO

    def get_file_storage(self) -> Storage:
        owner_field = self.get_owner_field()
        storage = getattr(owner_field, "storage", None) or default_storage
        if callable(storage):
            storage = storage()
        return storage

    def get_file(self) -> FieldFile:
        return self.file

    @classmethod
    def get_file_field(cls) -> VariationalFileField:
        return cls._meta.get_field("file")

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "file_info": "({ext}, {width}x{height}, {size})".format(
                ext=self.extension,
                width=self.width,
                height=self.height,
                size=filesizeformat(self.size),
            ),
        }

    @cached_method("_variations_cache")
    def get_variations(self) -> Dict[str, PaperVariation]:
        owner_field = self.get_owner_field()
        if owner_field is None:
            raise cached_method.Bypass({})

        variation_config = getattr(owner_field, "variations", {}).copy()
        return build_variations(variation_config)

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }


class UploadedImage(UploadedImageBase):
    class Meta(UploadedImageBase.Meta):
        pass



from django import forms
from django.utils.translation import gettext_lazy as _
from paper_admin.admin.renderers import PaperFormRenderer


class ChangeFileResourceDialogBase(forms.ModelForm):
    default_renderer = PaperFormRenderer
    new_name = forms.CharField(
        required=True,
        label=_("File name"),
        max_length=255,
    )

    class Meta:
        fields = ["new_name"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance:
            self.fields["new_name"].initial = self.instance.resource_name

    def save(self, commit=True):
        old_name = self.instance.resource_name
        new_name = self.cleaned_data["new_name"]
        if old_name != new_name:
            if self.instance.extension:
                new_name += "." + self.instance.extension
            self.instance.rename_file(new_name)
        return super().save(commit)



from ...models.file import UploadedFile
from .base import ChangeFileResourceDialogBase


class ChangeUploadedFileDialog(ChangeFileResourceDialogBase):
    class Meta:
        model = UploadedFile
        fields = ["new_name", "display_name"]



import sys

import inquirer
import inquirer.themes as themes
from inquirer.render.console import ConsoleRender


class GreenPassion(themes.GreenPassion):
    def __init__(self):
        super().__init__()
        self.Checkbox.selection_icon = "⇨"
        self.Checkbox.selection_color = themes.term.bright_green
        self.Checkbox.selected_icon = "☑"
        self.Checkbox.selected_color = themes.term.bright_green
        self.Checkbox.unselected_icon = "☐"
        self.Checkbox.unselected_color = themes.term.normal
        self.List.selection_cursor = self.Checkbox.selection_icon
        self.List.selection_color = self.Checkbox.selection_color
        self.List.unselected_color = self.Checkbox.unselected_color


class CustomConsoleRender(ConsoleRender):
    def __init__(self, event_generator=None, theme=None, *args, **kwargs):
        theme = theme or GreenPassion()
        super().__init__(event_generator, theme, *args, **kwargs)

    @property
    def width(self):
        return 240

    def _print_header(self, render):
        base = render.get_header()

        header = base[: self.width - 9] + "..." if len(base) > self.width - 6 else base
        default_value = " ({color}{default}{normal})".format(
            default=render.question.default, color=self._theme.Question.default_color, normal=self.terminal.normal
        )
        show_default = render.question.default and render.show_default
        header += default_value if show_default else ""
        msg_template = (
            "{t.clear_eos}{tq.brackets_color}[{tq.mark_color}?{tq.brackets_color}]{t.normal} {msg}"
        )

        self.print_str(
            msg_template,
            msg=header,
            lf=not render.title_inline,
            tq=self._theme.Question,
        )

    def print_str(self, base, lf=False, **kwargs):
        message = base.format(t=self.terminal, **kwargs)

        if lf:
            self._position += message.count("\n") + 1

        print(message, end="\n" if lf else "")
        sys.stdout.flush()


def prompt_action(message, choices, **kwargs):
    """
    Запрос действия от пользователя из заданного списка.

    Пример:
        action = prompt_action(
            message="What do you want to do?",
            choices=["Run script", "Exit"]
        )

        if action == "Run script":
            ...
        elif action == "Exit":
            ...
    """
    default = kwargs.pop("default")
    kwargs.setdefault("carousel", True)

    return inquirer.prompt(
        [
            inquirer.List(
                "answer",
                message=message,
                choices=choices,
                default=default,
                **kwargs
            )
        ],
        render=CustomConsoleRender()
    )["answer"]


def prompt_variants(message, choices, **kwargs):
    """
    Запрос от пользователя множества вариантов из заданного списка.

    Пример:
        actions = prompt_variants(
            message="What fruits do you like?",
            choices=["Apple", "Banana", ""]
        )

        if "Apple" in actions:
            ...
        elif "Exit" in actions:
            ...
    """
    default = kwargs.pop("default")
    if isinstance(default, str):
        default = [default]

    kwargs.setdefault("carousel", True)

    while True:
        answer = inquirer.prompt(
            [
                inquirer.Checkbox(
                    "answer",
                    message=message,
                    choices=choices,
                    default=default,
                    **kwargs
                )
            ],
            render=CustomConsoleRender()
        )["answer"]

        if answer:
            return answer



from django.contrib import admin


class ResourceAdminBase(admin.ModelAdmin):
    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return False

    def log_addition(self, *args, **kwargs):
        return

    def log_change(self, *args, **kwargs):
        return

    def log_deletion(self, *args, **kwargs):
        return



from django.contrib import admin
from django.urls import path

from .. import views
from ..models.file import UploadedFile
from .base import ResourceAdminBase


class UploadedFileAdminBase(ResourceAdminBase):
    upload_view_class = views.file.UploadFileView
    delete_view_class = views.file.DeleteFileView
    change_view_class = views.file.ChangeFileView

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        urlpatterns = [
            path(
                "upload/",
                self.admin_site.admin_view(self.upload_view_class.as_view()),
                name="%s_%s_upload" % info,
            ),
            path(
                "delete/",
                self.admin_site.admin_view(self.delete_view_class.as_view()),
                name="%s_%s_delete" % info,
            ),
            path(
                "change/",
                self.admin_site.admin_view(self.change_view_class.as_view()),
                name="%s_%s_change" % info,
            ),
        ]
        return urlpatterns


@admin.register(UploadedFile)
class UploadedFileAdmin(UploadedFileAdminBase):
    pass



from django.contrib.admin import site
from django.contrib.admin.sites import NotRegistered
from django.urls import reverse_lazy

from ...helpers import iterate_parent_models
from .base import FileResourceWidgetBase
from .mixins import DisplayFileLimitationsMixin


class FileWidget(DisplayFileLimitationsMixin, FileResourceWidgetBase):
    template_name = "paper_uploads/widgets/file.html"

    class Media:
        css = {
            "all": [
                "paper_uploads/dist/widget.css"
            ]
        }
        js = [
            "paper_uploads/dist/widget.js",
        ]

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        for model_class in iterate_parent_models(self.model):
            if site.is_registered(model_class):
                context.update(self.get_extra_context(model_class))
                break
        else:
            raise NotRegistered("The model %s is not registered." % self.model.__name__)

        return context

    def get_extra_context(self, model):
        info = model._meta.app_label, model._meta.model_name
        return {
            "upload_url": reverse_lazy("admin:%s_%s_upload" % info),
            "change_url": reverse_lazy("admin:%s_%s_change" % info),
            "delete_url": reverse_lazy("admin:%s_%s_delete" % info),
        }
